/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XDP_HOST_DBUS_H__
#define __XDP_HOST_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.host.portal.Registry */

#define XDP_DBUS_HOST_TYPE_REGISTRY (xdp_dbus_host_registry_get_type ())
#define XDP_DBUS_HOST_REGISTRY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_HOST_TYPE_REGISTRY, XdpDbusHostRegistry))
#define XDP_DBUS_HOST_IS_REGISTRY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_HOST_TYPE_REGISTRY))
#define XDP_DBUS_HOST_REGISTRY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XDP_DBUS_HOST_TYPE_REGISTRY, XdpDbusHostRegistryIface))

struct _XdpDbusHostRegistry;
typedef struct _XdpDbusHostRegistry XdpDbusHostRegistry;
typedef struct _XdpDbusHostRegistryIface XdpDbusHostRegistryIface;

struct _XdpDbusHostRegistryIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_register) (
    XdpDbusHostRegistry *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_app_id,
    GVariant *arg_options);

  guint  (*get_version) (XdpDbusHostRegistry *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusHostRegistry, g_object_unref)
#endif

GType xdp_dbus_host_registry_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xdp_dbus_host_registry_interface_info (void);
guint xdp_dbus_host_registry_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xdp_dbus_host_registry_complete_register (
    XdpDbusHostRegistry *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void xdp_dbus_host_registry_call_register (
    XdpDbusHostRegistry *proxy,
    const gchar *arg_app_id,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_dbus_host_registry_call_register_finish (
    XdpDbusHostRegistry *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xdp_dbus_host_registry_call_register_sync (
    XdpDbusHostRegistry *proxy,
    const gchar *arg_app_id,
    GVariant *arg_options,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
guint xdp_dbus_host_registry_get_version (XdpDbusHostRegistry *object);
void xdp_dbus_host_registry_set_version (XdpDbusHostRegistry *object, guint value);


/* ---- */

#define XDP_DBUS_HOST_TYPE_REGISTRY_PROXY (xdp_dbus_host_registry_proxy_get_type ())
#define XDP_DBUS_HOST_REGISTRY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_HOST_TYPE_REGISTRY_PROXY, XdpDbusHostRegistryProxy))
#define XDP_DBUS_HOST_REGISTRY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_DBUS_HOST_TYPE_REGISTRY_PROXY, XdpDbusHostRegistryProxyClass))
#define XDP_DBUS_HOST_REGISTRY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_DBUS_HOST_TYPE_REGISTRY_PROXY, XdpDbusHostRegistryProxyClass))
#define XDP_DBUS_HOST_IS_REGISTRY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_HOST_TYPE_REGISTRY_PROXY))
#define XDP_DBUS_HOST_IS_REGISTRY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_DBUS_HOST_TYPE_REGISTRY_PROXY))

typedef struct _XdpDbusHostRegistryProxy XdpDbusHostRegistryProxy;
typedef struct _XdpDbusHostRegistryProxyClass XdpDbusHostRegistryProxyClass;
typedef struct _XdpDbusHostRegistryProxyPrivate XdpDbusHostRegistryProxyPrivate;

struct _XdpDbusHostRegistryProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XdpDbusHostRegistryProxyPrivate *priv;
};

struct _XdpDbusHostRegistryProxyClass
{
  GDBusProxyClass parent_class;
};

GType xdp_dbus_host_registry_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusHostRegistryProxy, g_object_unref)
#endif

void xdp_dbus_host_registry_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpDbusHostRegistry *xdp_dbus_host_registry_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XdpDbusHostRegistry *xdp_dbus_host_registry_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xdp_dbus_host_registry_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpDbusHostRegistry *xdp_dbus_host_registry_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XdpDbusHostRegistry *xdp_dbus_host_registry_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON (xdp_dbus_host_registry_skeleton_get_type ())
#define XDP_DBUS_HOST_REGISTRY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON, XdpDbusHostRegistrySkeleton))
#define XDP_DBUS_HOST_REGISTRY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON, XdpDbusHostRegistrySkeletonClass))
#define XDP_DBUS_HOST_REGISTRY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON, XdpDbusHostRegistrySkeletonClass))
#define XDP_DBUS_HOST_IS_REGISTRY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON))
#define XDP_DBUS_HOST_IS_REGISTRY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_DBUS_HOST_TYPE_REGISTRY_SKELETON))

typedef struct _XdpDbusHostRegistrySkeleton XdpDbusHostRegistrySkeleton;
typedef struct _XdpDbusHostRegistrySkeletonClass XdpDbusHostRegistrySkeletonClass;
typedef struct _XdpDbusHostRegistrySkeletonPrivate XdpDbusHostRegistrySkeletonPrivate;

struct _XdpDbusHostRegistrySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XdpDbusHostRegistrySkeletonPrivate *priv;
};

struct _XdpDbusHostRegistrySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xdp_dbus_host_registry_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusHostRegistrySkeleton, g_object_unref)
#endif

XdpDbusHostRegistry *xdp_dbus_host_registry_skeleton_new (void);


G_END_DECLS

#endif /* __XDP_HOST_DBUS_H__ */
