/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GEOCLUE_DBUS_H__
#define __GEOCLUE_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.GeoClue2.Client */

#define GEOCLUE_TYPE_CLIENT (geoclue_client_get_type ())
#define GEOCLUE_CLIENT(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GEOCLUE_TYPE_CLIENT, GeoclueClient))
#define GEOCLUE_IS_CLIENT(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GEOCLUE_TYPE_CLIENT))
#define GEOCLUE_CLIENT_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GEOCLUE_TYPE_CLIENT, GeoclueClientIface))

struct _GeoclueClient;
typedef struct _GeoclueClient GeoclueClient;
typedef struct _GeoclueClientIface GeoclueClientIface;

struct _GeoclueClientIface
{
  GTypeInterface parent_iface;



  gboolean (*handle_start) (
    GeoclueClient *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_stop) (
    GeoclueClient *object,
    GDBusMethodInvocation *invocation);

  gboolean  (*get_active) (GeoclueClient *object);

  const gchar * (*get_desktop_id) (GeoclueClient *object);

  guint  (*get_distance_threshold) (GeoclueClient *object);

  const gchar * (*get_location) (GeoclueClient *object);

  guint  (*get_requested_accuracy_level) (GeoclueClient *object);

  guint  (*get_time_threshold) (GeoclueClient *object);

  void (*location_updated) (
    GeoclueClient *object,
    const gchar *arg_old,
    const gchar *arg_new);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GeoclueClient, g_object_unref)
#endif

GType geoclue_client_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *geoclue_client_interface_info (void);
guint geoclue_client_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void geoclue_client_complete_start (
    GeoclueClient *object,
    GDBusMethodInvocation *invocation);

void geoclue_client_complete_stop (
    GeoclueClient *object,
    GDBusMethodInvocation *invocation);



/* D-Bus signal emissions functions: */
void geoclue_client_emit_location_updated (
    GeoclueClient *object,
    const gchar *arg_old,
    const gchar *arg_new);



/* D-Bus method calls: */
void geoclue_client_call_start (
    GeoclueClient *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean geoclue_client_call_start_finish (
    GeoclueClient *proxy,
    GAsyncResult *res,
    GError **error);

gboolean geoclue_client_call_start_sync (
    GeoclueClient *proxy,
    GCancellable *cancellable,
    GError **error);

void geoclue_client_call_stop (
    GeoclueClient *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean geoclue_client_call_stop_finish (
    GeoclueClient *proxy,
    GAsyncResult *res,
    GError **error);

gboolean geoclue_client_call_stop_sync (
    GeoclueClient *proxy,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
const gchar *geoclue_client_get_location (GeoclueClient *object);
gchar *geoclue_client_dup_location (GeoclueClient *object);
void geoclue_client_set_location (GeoclueClient *object, const gchar *value);

guint geoclue_client_get_distance_threshold (GeoclueClient *object);
void geoclue_client_set_distance_threshold (GeoclueClient *object, guint value);

guint geoclue_client_get_time_threshold (GeoclueClient *object);
void geoclue_client_set_time_threshold (GeoclueClient *object, guint value);

const gchar *geoclue_client_get_desktop_id (GeoclueClient *object);
gchar *geoclue_client_dup_desktop_id (GeoclueClient *object);
void geoclue_client_set_desktop_id (GeoclueClient *object, const gchar *value);

guint geoclue_client_get_requested_accuracy_level (GeoclueClient *object);
void geoclue_client_set_requested_accuracy_level (GeoclueClient *object, guint value);

gboolean geoclue_client_get_active (GeoclueClient *object);
void geoclue_client_set_active (GeoclueClient *object, gboolean value);


/* ---- */

#define GEOCLUE_TYPE_CLIENT_PROXY (geoclue_client_proxy_get_type ())
#define GEOCLUE_CLIENT_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GEOCLUE_TYPE_CLIENT_PROXY, GeoclueClientProxy))
#define GEOCLUE_CLIENT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GEOCLUE_TYPE_CLIENT_PROXY, GeoclueClientProxyClass))
#define GEOCLUE_CLIENT_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GEOCLUE_TYPE_CLIENT_PROXY, GeoclueClientProxyClass))
#define GEOCLUE_IS_CLIENT_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GEOCLUE_TYPE_CLIENT_PROXY))
#define GEOCLUE_IS_CLIENT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GEOCLUE_TYPE_CLIENT_PROXY))

typedef struct _GeoclueClientProxy GeoclueClientProxy;
typedef struct _GeoclueClientProxyClass GeoclueClientProxyClass;
typedef struct _GeoclueClientProxyPrivate GeoclueClientProxyPrivate;

struct _GeoclueClientProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GeoclueClientProxyPrivate *priv;
};

struct _GeoclueClientProxyClass
{
  GDBusProxyClass parent_class;
};

GType geoclue_client_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GeoclueClientProxy, g_object_unref)
#endif

void geoclue_client_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GeoclueClient *geoclue_client_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GeoclueClient *geoclue_client_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void geoclue_client_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GeoclueClient *geoclue_client_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GeoclueClient *geoclue_client_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GEOCLUE_TYPE_CLIENT_SKELETON (geoclue_client_skeleton_get_type ())
#define GEOCLUE_CLIENT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GEOCLUE_TYPE_CLIENT_SKELETON, GeoclueClientSkeleton))
#define GEOCLUE_CLIENT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GEOCLUE_TYPE_CLIENT_SKELETON, GeoclueClientSkeletonClass))
#define GEOCLUE_CLIENT_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GEOCLUE_TYPE_CLIENT_SKELETON, GeoclueClientSkeletonClass))
#define GEOCLUE_IS_CLIENT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GEOCLUE_TYPE_CLIENT_SKELETON))
#define GEOCLUE_IS_CLIENT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GEOCLUE_TYPE_CLIENT_SKELETON))

typedef struct _GeoclueClientSkeleton GeoclueClientSkeleton;
typedef struct _GeoclueClientSkeletonClass GeoclueClientSkeletonClass;
typedef struct _GeoclueClientSkeletonPrivate GeoclueClientSkeletonPrivate;

struct _GeoclueClientSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GeoclueClientSkeletonPrivate *priv;
};

struct _GeoclueClientSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType geoclue_client_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GeoclueClientSkeleton, g_object_unref)
#endif

GeoclueClient *geoclue_client_skeleton_new (void);


G_END_DECLS

#endif /* __GEOCLUE_DBUS_H__ */
