/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XDP_BACKGROUND_DBUS_H__
#define __XDP_BACKGROUND_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.background.Monitor */

#define XDP_DBUS_BACKGROUND_TYPE_MONITOR (xdp_dbus_background_monitor_get_type ())
#define XDP_DBUS_BACKGROUND_MONITOR(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR, XdpDbusBackgroundMonitor))
#define XDP_DBUS_BACKGROUND_IS_MONITOR(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR))
#define XDP_DBUS_BACKGROUND_MONITOR_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR, XdpDbusBackgroundMonitorIface))

struct _XdpDbusBackgroundMonitor;
typedef struct _XdpDbusBackgroundMonitor XdpDbusBackgroundMonitor;
typedef struct _XdpDbusBackgroundMonitorIface XdpDbusBackgroundMonitorIface;

struct _XdpDbusBackgroundMonitorIface
{
  GTypeInterface parent_iface;

  GVariant * (*get_background_apps) (XdpDbusBackgroundMonitor *object);

  guint  (*get_version) (XdpDbusBackgroundMonitor *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusBackgroundMonitor, g_object_unref)
#endif

GType xdp_dbus_background_monitor_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xdp_dbus_background_monitor_interface_info (void);
guint xdp_dbus_background_monitor_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus property accessors: */
GVariant *xdp_dbus_background_monitor_get_background_apps (XdpDbusBackgroundMonitor *object);
GVariant *xdp_dbus_background_monitor_dup_background_apps (XdpDbusBackgroundMonitor *object);
void xdp_dbus_background_monitor_set_background_apps (XdpDbusBackgroundMonitor *object, GVariant *value);

guint xdp_dbus_background_monitor_get_version (XdpDbusBackgroundMonitor *object);
void xdp_dbus_background_monitor_set_version (XdpDbusBackgroundMonitor *object, guint value);


/* ---- */

#define XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY (xdp_dbus_background_monitor_proxy_get_type ())
#define XDP_DBUS_BACKGROUND_MONITOR_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY, XdpDbusBackgroundMonitorProxy))
#define XDP_DBUS_BACKGROUND_MONITOR_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY, XdpDbusBackgroundMonitorProxyClass))
#define XDP_DBUS_BACKGROUND_MONITOR_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY, XdpDbusBackgroundMonitorProxyClass))
#define XDP_DBUS_BACKGROUND_IS_MONITOR_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY))
#define XDP_DBUS_BACKGROUND_IS_MONITOR_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_DBUS_BACKGROUND_TYPE_MONITOR_PROXY))

typedef struct _XdpDbusBackgroundMonitorProxy XdpDbusBackgroundMonitorProxy;
typedef struct _XdpDbusBackgroundMonitorProxyClass XdpDbusBackgroundMonitorProxyClass;
typedef struct _XdpDbusBackgroundMonitorProxyPrivate XdpDbusBackgroundMonitorProxyPrivate;

struct _XdpDbusBackgroundMonitorProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XdpDbusBackgroundMonitorProxyPrivate *priv;
};

struct _XdpDbusBackgroundMonitorProxyClass
{
  GDBusProxyClass parent_class;
};

GType xdp_dbus_background_monitor_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusBackgroundMonitorProxy, g_object_unref)
#endif

void xdp_dbus_background_monitor_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpDbusBackgroundMonitor *xdp_dbus_background_monitor_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XdpDbusBackgroundMonitor *xdp_dbus_background_monitor_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xdp_dbus_background_monitor_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpDbusBackgroundMonitor *xdp_dbus_background_monitor_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XdpDbusBackgroundMonitor *xdp_dbus_background_monitor_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON (xdp_dbus_background_monitor_skeleton_get_type ())
#define XDP_DBUS_BACKGROUND_MONITOR_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON, XdpDbusBackgroundMonitorSkeleton))
#define XDP_DBUS_BACKGROUND_MONITOR_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON, XdpDbusBackgroundMonitorSkeletonClass))
#define XDP_DBUS_BACKGROUND_MONITOR_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON, XdpDbusBackgroundMonitorSkeletonClass))
#define XDP_DBUS_BACKGROUND_IS_MONITOR_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON))
#define XDP_DBUS_BACKGROUND_IS_MONITOR_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_DBUS_BACKGROUND_TYPE_MONITOR_SKELETON))

typedef struct _XdpDbusBackgroundMonitorSkeleton XdpDbusBackgroundMonitorSkeleton;
typedef struct _XdpDbusBackgroundMonitorSkeletonClass XdpDbusBackgroundMonitorSkeletonClass;
typedef struct _XdpDbusBackgroundMonitorSkeletonPrivate XdpDbusBackgroundMonitorSkeletonPrivate;

struct _XdpDbusBackgroundMonitorSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XdpDbusBackgroundMonitorSkeletonPrivate *priv;
};

struct _XdpDbusBackgroundMonitorSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xdp_dbus_background_monitor_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpDbusBackgroundMonitorSkeleton, g_object_unref)
#endif

XdpDbusBackgroundMonitor *xdp_dbus_background_monitor_skeleton_new (void);


G_END_DECLS

#endif /* __XDP_BACKGROUND_DBUS_H__ */
